/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef REGISTERS_H_
#define REGISTERS_H_

#include "types.h"

enum Registers_Bank1 {
    VER             = 0x00,
    DSID            = 0x02,
    FID             = 0x06,
    VID             = 0x08,
    PID             = 0x0a,
    ADR             = 0x0c,
    VSK             = 0x0e,
    GRP             = 0x10,
    RebootCnt       = 0x1e,
    SensorTableStart      = 0x20,
    SensorTableLast       = 0x40,
    BinaryInputTableStart = 0x40,
    BinaryInputTableLast  = 0x58,
    OpcTableStart         = 0x80,
};

enum Registers_Bank3 {
    MODE            = 0x00,
    BUTTON_ID_GRP_0 = 0x01,
    PBCHANNEL       = 0x1c,
    PBGROUP         = 0x1d,
    LTMODE          = 0x1e,
};

#define MODEHasOutputLoad(mode_)   ((mode_) >= 16)

#define FidGetClass(fid_)           ((fid_) >> 12)
#define FidGetSubclass(fid_)        (((fid_) >> 6) & 0x3F)

/**
 * FID Class
 */
enum FidClass {
    FID_CLASS_OVERALL   = 0,
    FID_CLASS_LIGHT     = 1,
    FID_CLASS_SHADE     = 2,
    FID_CLASS_CLIMATE   = 3,
    FID_CLASS_AUDIO     = 4,
    FID_CLASS_VIDEO     = 5,
    FID_CLASS_SECURITY  = 6,
    FID_CLASS_ACCESS    = 7,
    FID_CLASS_JOKER     = 8,
    FID_CLASS_WHITE     = 9,
};

enum FidSubclass {
    FID_SUBCLASS_STANDARD_R100      = 0x00,
    FID_SUBCLASS_COLOR_SCENE_DEVICE = 0x00,
    FID_SUBCLASS_1WAY_R105          = 0x04,
    FID_SUBCLASS_SCENE_DEVICE       = 0x05,
    FID_SUBCLASS_PRE_R200           = 0x06,
    FID_SUBCLASS_CUSTOM_DEVICE      = 0x3F,
};

enum FunctionModuleR100 {
    /* output */
    FM_OUTPUT_MASK          = 0x10,     /**< output mask */
    FM_OUTPUT_NONE          = 0x00,     /**< no output */
    FM_OUTPUT_SWITCH        = 0x10,     /**< output is a switch */
    FM_VAR_BTN_GRP_MASK     = 0x20,     /**< variable button configuration mask */
    FM_VAR_BTN_GRP_CFG      = 0x20,     /**< variable button configuration */
    /* button configuration */
    FM_BUTTON_CFG_MASK      = 0x07,
    FM_BUTTON_CFG_1X        = 0x00,
    FM_BUTTON_CFG_2X        = 0x01,
    FM_BUTTON_CFG_4X        = 0x02,
    FM_BUTTON_CFG_NONE      = 0x07,
};

enum FunctionModuleR105 {
    FM_VARIABLE_OUTPUT_SLOPE = 0x20,    /**< Variable output slope */
    FM_OUTPUT_PRESENT        = 0x10,    /**< Device has load output */
    FM_ADDITONAL_HW          = 0x08,    /**< Device has additional capabilities */
    FM_VARIABLE_STD_GROUP    = 0x04,    /**< Device has variable standard group */
    FM_R105_BUTTON_MASK      = 0x03,
    FM_R105_BUTTON_CFG_0X    = 0x00,
    FM_R105_BUTTON_CFG_1X    = 0x01,
    FM_R105_BUTTON_CFG_2X    = 0x02,
    FM_R105_BUTTON_CFG_4X    = 0x03,
};

enum FunctionModulePreR200 {
    FM_PRE_R200_OUTPUT       = 0x01,    /**< Device has output(s) */
    FM_PRE_R200_BINARY_INPUT = 0x02,    /**< Device has binary input(s) */
    FM_PRE_R200_SENSOR       = 0x04,    /**< Device has sensor(s) */
    FM_PRE_R200_BUTTON       = 0x08,    /**< Device has button(s) */
};

/**
 * Function Module 9-Way Button
 */
enum FunctionModule9Way {
    /* sensor */
    FM_9WAY_SENSOR_MASK     = 0x0C,
    FM_9WAY_SENSOR          = 0x04,
    /* button */
    FM_9WAY_BUTTON_MASK     = 0x03,
    FM_9WAY_BUTTON_9WAY     = 0x00
};

/**
 * FID Access Subclasses
 */
enum FidSubclassAccess {
    FID_SUBCLASS_ACCESS_BELL    = 0x00,
    FID_SUBCLASS_ACCESS_LEAVING = 0x01,
    FID_SUBCLASS_ACCESS_LOCKING = 0x02,
};

/**
 * Function Module Access: Bell
 */
enum FunctionModuleAccessBell {
    /* local button */
    FM_BELL_LOCAL_BUTTON_MASK  = 0x08,
    FM_BELL_LOCAL_BUTTON_NONE  = 0x00,
    FM_BELL_LOCAL_BUTTON_AVAIL = 0x08,
    /* button */
    FM_BELL_BUTTON_MASK        = 0x07,
    FM_BELL_BUTTON_1WAY        = 0x00,
    FM_BELL_BUTTON_2WAY        = 0x01,
    FM_BELL_BUTTON_4WAY        = 0x03,
    FM_BELL_BUTTON_NONE        = 0x07,
};

/**
 * Function Module Access: Come/Leave
 */
enum FunctionModuleAccessLeave {
    /* button */
    FM_LEAVE_BUTTON_MASK        = 0x07,
    FM_LEAVE_BUTTON_LEAVE       = 0x00,
    FM_LEAVE_BUTTON_COME        = 0x01,
    FM_LEAVE_BUTTON_LEAVE_COME  = 0x02, /* Button1 = LEAVE; Button2 = COME */
};

/**
 * FID Security Subclasses
 */
enum FidSubclassSecurity {
    FID_SUBCLASS_SECURITY_ALARM = 0x00,
};

/**
 * Function Module Security: Alarm
 */
enum FunctionModuleSecurityAlarm {
    /* button */
    FM_ALARM_BUTTON_MASK        = 0x03,
    FM_ALARM_BUTTON_ALARM       = 0x00,
    FM_ALARM_BUTTON_PANIC       = 0x01,
    FM_ALARM_BUTTON_ALARM_PANIC = 0x02, /* Button1 = ALARM; Button2 = PANIC */
};

/**
 * Function Module Climate: Custom
 */
enum FunctionModuleClimateCustom {
    /* button */
    FM_CLIMATE_CUSTOM_BS2       = 0x00,
};

bool Fid_IsGroupMatching(uint16_t fid, uint8_t group);
bool Fid_IsdSLinkDevice(uint16_t fid);

bool Fid_HasButton(uint16_t fid);
bool Fid_HasVariableButtonIdGroup(uint16_t fid);
bool Fid_HasOutput(uint16_t fid);
bool Fid_HasBinaryInputs(uint16_t fid);
bool Fid_HasSensors(uint16_t fid);
bool Fid_HasOutputChannels(uint16_t fid);
bool Fid_HasButtonOutputChannel(uint16_t fid);
bool Fid_HasVariableStandardGroup(uint16_t fid);

bool Fid_IsSecurityAlarmButton(uint16_t fid, uint8_t key);
bool Fid_IsSecurityPanicButton(uint16_t fid, uint8_t key);

bool Fid_IsAccessLeaveButton(uint16_t fid, uint8_t key);

bool Fid_IsJocker(uint16_t fid);

#endif /* REGISTERS_H_ */
